@@ -17,6 +17,11 @@ import six
 import pkg_resources
 import requests
 
+try:
+    from bs4 import BeautifulSoup
+except ImportError:
+    BeautifulSoup = None
+
 try:
     import argcomplete
 except ImportError:
@@ -25,10 +30,13 @@ except ImportError:
 import copr.exceptions as copr_exceptions
 from copr.v3 import (
     Client, config_from_file, CoprException, CoprRequestException,
-    CoprNoConfigException, CoprConfigException, CoprNoResultException,
+    CoprConfigException, CoprNoResultException, CoprAuthException,
 )
 from copr.v3.pagination import next_page
-from .util import ProgressBar, json_dumps, serializable
+from copr_cli.helpers import cli_use_output_format
+from copr_cli.monitor import cli_monitor_parser
+from copr_cli.printers import cli_get_output_printer as get_printer
+from copr_cli.util import get_progress_callback, serializable
 from .build_config import MockProfile
 
 
@@ -37,15 +45,7 @@ if six.PY2:
 else:
     from urllib.parse import urljoin, urlparse
 
-if sys.version_info < (2, 7):
-    class NullHandler(logging.Handler):
-        def emit(self, record):
-            pass
-else:
-    from logging import NullHandler
-
 log = logging.getLogger(__name__)
-log.addHandler(NullHandler())
 
 
 ON_OFF_MAP = {
@@ -62,24 +62,6 @@ BOOTSTRAP_MAP = {
     None: "default",
 }
 
-no_config_warning = """
-================= WARNING: =======================
-File '{0}' is missing or incorrect.
-See documentation: man copr-cli.
-Any operation requiring credentials will fail!
-==================================================
-
-Hint: {1}
-
-"""
-
-output_format_help = """
-Set the formatting style. We recommend using json, which prints the required data
-in json format. The text format prints the required data in a column, one piece of
-information per line. The text-row format prints all information separated
-by a space on a single line.
-"""
-
 try:
     input = raw_input
 except NameError:
@@ -110,7 +92,7 @@ class ActionDeprecated(argparse.Action):
         setattr(namespace, self.dest, values)
 
 
-def buildopts_from_args(args, progress_callback):
+def buildopts_from_args(args, progress_callback=None):
     """
     For all the build commands, parse the common set of build options.
     """
@@ -131,177 +113,69 @@ def buildopts_from_args(args, progress_callback):
     return buildopts
 
 
-class AbstractPrinter(object):
-    """Abstract class defining mandatory methods of printer classes"""
-
-    def __init__(self, fields):
-        """Represents the data we want to print.
-        Supports callable lambda function which takes exactly one argument"""
-        self.fields = fields
-
-    def add_data(self, data):
-        """Initialize the data to be printed"""
-        raise NotImplementedError
-
-    def finish(self):
-        """Print the data according to the set format"""
-        raise NotImplementedError
-
-
-class RowTextPrinter(AbstractPrinter):
-    """The class takes care of printing the data in row text format"""
-
-    def finish(self):
-        pass
-
-    def add_data(self, data):
-        row_data = []
-        for field in self.fields:
-            if callable(field):
-                row_data.append(str(field(data)))
-            else:
-                row_data.append(str(data[field]))
-        print("\t".join(row_data))
-
-
-class ColumnTextPrinter(AbstractPrinter):
-    """The class takes care of printing the data in column text format"""
-
-    first_line = True
-
-    def finish(self):
-        pass
-
-    def add_data(self, data):
-        if not self.first_line:
-            print()
-        self.first_line = False
-        for field in self.fields:
-            if callable(field):
-                print("{0}: {1}".format(field.__code__.co_varnames[0], str(field(data))))
-            else:
-                print("{0}: {1}".format(field, str(data[field])))
-
-
-class JsonPrinter(AbstractPrinter):
-    """The class takes care of printing the data in json format"""
-
-    def _get_result_json(self, data):
-        result = {}
-        for field in self.fields:
-            if callable(field):
-                name = field.__code__.co_varnames[0]
-                result[name] = field(data)
-            else:
-                result[field] = data[field]
-        return json_dumps(result)
-
-    def add_data(self, data):
-        print(self._get_result_json(data))
-
-    def finish(self):
-        pass
-
-
-class JsonPrinterListCommand(JsonPrinter):
+def requires_api_auth(func):
     """
-    The class takes care of printing the data in list in json format
-
-    For performance reasons we cannot simply add everything to a list and then
-    use `json_dumps` function to print all JSON at once. For large projects this
-    would mean minutes of no output, which is not user-friendly.
-
-    The approach here is to utilize `json_dumps` to convert each object to
-    a JSON string and print it immediately. We then manually take care of
-    opening and closing list brackets and separators.
+    For operations that need authentication with server, check early that we
+    have either the user/token pair in hand, or that we can use the GSSAPI
+    session for authentication.  IOW, this decorator contacts the server
+    only if the GSSAPI is ON and preferred -- and the session cookie is not
+    initialized.
     """
+    def wrapper(self, args):
+        # This call implies that either login+token is specified, GSSAPI
+        # session file exists, or (as a fallback) that we are able to do
+        # GSSAPI handshake and initialize the GSSAPI session.
+        self.client.base_proxy.auth_username()
+        return func(self, args)
 
-    first_line = True
-
-    def add_data(self, data):
-        self.start()
-        result = self._get_result_json(data)
-        for line in result.split("\n"):
-            print("    {0}".format(line))
-
-    def start(self):
-        """
-        This is called before printing the first object
-        """
-        if self.first_line:
-            self.first_line = False
-            print("[")
-        else:
-            print("    ,")
-
-    def finish(self):
-        """
-        This is called by the user after printing all objects
-        """
-        if not self.first_line:
-            print("]")
-
-
-def get_printer(output_format, fields, list_command=False):
-    """According to output_format decide which object of printer to return"""
-    if output_format == "json":
-        if list_command:
-            return JsonPrinterListCommand(fields)
-        return JsonPrinter(fields)
-    if output_format == "text":
-        return ColumnTextPrinter(fields)
-    return RowTextPrinter(fields)
+    wrapper.__doc__ = func.__doc__
+    wrapper.__name__ = func.__name__
+    return wrapper
 
 
 class Commands(object):
     def __init__(self, config_path):
-        self.config_path = config_path or '~/.config/copr'
-
-        try:
-            self.config = config_from_file(self.config_path)
-        except (CoprNoConfigException, CoprConfigException) as ex:
-            sys.stderr.write(no_config_warning.format(self.config_path, ex))
-            self.config = {"copr_url": "http://copr.fedoraproject.org", "no_config": True}
+        self.config = config_from_file(config_path)
 
+        if self.config.get("gssapi") is None:
+            # Contrary to what is set in python-copr, we set GSSAPI on by
+            # default (unless user explicitly says otherwise).
+            self.config["gssapi"] = True
+        self.config["connection_attempts"] = 3
         self.client = Client(self.config)
 
-    def requires_api_auth(func):
-        """ Decorator that checks config presence
+    @property
+    def username(self):
         """
-
-        def wrapper(self, args):
-            if "no_config" in self.config:
-                sys.stderr.write("Error: Operation requires api authentication\n")
-                sys.exit(6)
-
-            return func(self, args)
-
-        wrapper.__doc__ = func.__doc__
-        wrapper.__name__ = func.__name__
-        return wrapper
-
-    def check_username_presence(func):
-        """ Decorator that checks if username was provided
+        Get the username from config, or obtain it via auth_check (transitively
+        via GSSAPI).
         """
+        if self.config.get("username"):
+            return self.config["username"]
 
-        def wrapper(self, args):
-            if "no_config" in self.config and args.username is None:
-                sys.stderr.write(
-                    "Error: Operation requires username\n"
-                    "Pass username to command or create `{0}`\n".format(self.config_path))
-                sys.exit(6)
+        if self.config.get("gssapi"):
+            try:
+                return self.client.base_proxy.auth_username()
+            except CoprAuthException:
+                log.error("Failed to determine Copr username from authentication.")
+                raise
+        raise CoprConfigException(
+            "This operation tries to detect your username, but it is not "
+            "possible to find it in configuration, and GSSAPI is disabled "
+        )
 
-            if args.username is None and self.config["username"] is None:
-                sys.stderr.write(
-                    "Error: Operation requires username\n"
-                    "Pass username to command or add it to `{0}`\n".format(self.config_path))
-                sys.exit(6)
-
-            return func(self, args)
-
-        wrapper.__doc__ = func.__doc__
-        wrapper.__name__ = func.__name__
-        return wrapper
+    @property
+    def ownername(self):
+        """
+        Determine the project ownername (== username) when not specified on
+        commandline.
+        """
+        try:
+            return self.username
+        except:
+            log.error("This operation needs a project ownername specified, "
+                      "fallback to your username failed.")
+            raise
 
     def parse_name(self, name):
         m = re.match(r"([^/]+)/(.*)", name)
@@ -309,9 +183,27 @@ class Commands(object):
             owner = m.group(1)
             name = m.group(2)
         else:
-            owner = self.config["username"]
+            owner = self.ownername
         return owner, name
 
+    def parse_dirname(self, name):
+        """
+        From specified <owner>/<dirname> or just <dirname> return the
+        tripled (owner, project, dirname).
+        """
+        owner, dirname = self.parse_name(name)
+        project = dirname.split(':')[0]
+        return owner, project, dirname
+
+
+    def build_url(self, build_id):
+        """
+        Return the "generic" predictable url for build_id, which redirects
+        to the final owner/project/build_id route.
+        """
+        return urljoin(self.config["copr_url"],
+                       "/coprs/build/{0}".format(build_id))
+
     def parse_chroot_path(self, path):
         """
         Take a `path` in an `owner/project/chroot` format and return a tuple of
@@ -319,7 +211,7 @@ class Commands(object):
         """
         m = re.match(r"(([^/]+)/)?([^/]+)/(.*)", path)
         if m:
-            owner = m.group(2) or self.config["username"]
+            owner = m.group(2) or self.username
             return owner, m.group(3), m.group(4)
         raise CoprException("Unexpected chroot path format")
 
@@ -331,6 +223,7 @@ class Commands(object):
 
         prevstatus = defaultdict(lambda: None)
         failed_ids = []
+        canceled_ids = []
 
         watched = set(build_ids)
         done = set()
@@ -354,6 +247,8 @@ class Commands(object):
 
                     if build_details.state in ["failed"]:
                         failed_ids.append(build_id)
+                    if build_details.state in ["canceled"]:
+                        canceled_ids.append(build_id)
                     if build_details.state in ["succeeded", "skipped",
                                                "failed", "canceled"]:
                         done.add(build_id)
@@ -366,19 +261,25 @@ class Commands(object):
 
                 time.sleep(30)
 
+            exception_message = ""
+            separator = ""
             if failed_ids:
-                raise copr_exceptions.CoprBuildException(
-                    "Build(s) {0} failed.".format(
-                        ", ".join(str(x) for x in failed_ids)))
+                exception_message = "Build(s) {0} failed.".format(", ".join(str(x) for x in failed_ids))
+                separator = " "
+            if canceled_ids:
+                exception_message += separator + "Build(s) {0} canceled.".format(", ".join(str(x) for x in canceled_ids))
+
+            if failed_ids or canceled_ids:
+                raise copr_exceptions.CoprBuildException(exception_message)
 
         except KeyboardInterrupt:
             pass
 
-    def action_whoami(self, args):
+    def action_whoami(self, _args):
         """
         Simply print out the current user as defined in copr config.
         """
-        print(self.config["username"])
+        print(self.username)
 
     def action_new_webhook_secret(self, args):
         """
@@ -406,31 +307,39 @@ class Commands(object):
 
         :param args: argparse arguments provided by the user
         """
-        self.client.build_proxy.auth_check()
+
+        # Before we start uploading potentially large source RPM file, make sure
+        # that the user actually has a valid credentials.
+        self.client.base_proxy.auth_check()
+
+        username, projectname, project_dirname = self.parse_dirname(args.copr_repo)
 
         builds = []
         for pkg in args.pkgs:
             if os.path.exists(pkg):
-                bar = ProgressBar(max=os.path.getsize(pkg))
-                build_function = self.client.build_proxy.create_from_file
+                progress_callback = get_progress_callback(os.path.getsize(pkg))
+                buildopts = buildopts_from_args(args, progress_callback)
                 data = {"path": pkg}
-
-                # pylint: disable=function-redefined
-                def progress_callback(monitor):
-                    bar.next(n=8192)
-
                 print('Uploading package {0}'.format(pkg))
+                try:
+                    builds.append(self.client.build_proxy.create_from_file(
+                        ownername=username, projectname=projectname,
+                        project_dirname=project_dirname, buildopts=buildopts,
+                        **data))
+                finally:
+                    if progress_callback:
+                        progress_callback.finish()
             elif not urlparse(pkg).scheme:
                 raise CoprException("File {0} not found".format(pkg))
             else:
-                bar = None
-                progress_callback = None
-                build_function = self.client.build_proxy.create_from_url
+                buildopts = buildopts_from_args(args)
                 data = {"url": pkg}
+                builds.append(self.client.build_proxy.create_from_url(
+                    ownername=username, projectname=projectname,
+                    project_dirname=project_dirname, buildopts=buildopts,
+                    **data))
 
-            builds.append(self.process_build(args, build_function, data, bar=bar, progress_callback=progress_callback))
-
-        return builds
+        self.print_build_info_and_wait(builds, args)
 
     @requires_api_auth
     def action_build_pypi(self, args):
@@ -442,6 +351,7 @@ class Commands(object):
         data = {
             "pypi_package_name": args.packagename,
             "pypi_package_version": args.packageversion,
+            "spec_generator": args.spec_generator,
             "spec_template": args.spec_template,
             "python_versions": args.pythonversions,
         }
@@ -501,44 +411,34 @@ class Commands(object):
             data[arg] = getattr(args, arg)
         return self.process_build(args, self.client.build_proxy.create_from_custom, data)
 
-    @requires_api_auth
-    def action_build_distgit(self, args):
+    def print_build_info_and_wait(self, builds, args):
         """
-        Method called when the 'buildfedpkg' action has been selected by the user.
-
-        :param args: argparse arguments provided by the user
+        Print to stdout information about just submitted builds.  Note that our
+        users parse this non-parseable output, because we don't have issue #224
+        fixed.
         """
-        data = {"clone_url": args.clone_url, "committish": args.branch}
-        return self.process_build(args, self.client.build_proxy.create_from_scm, data)
 
-    def process_build(self, args, build_function, data, bar=None, progress_callback=None):
-        username, project_dirname = self.parse_name(args.copr_repo)
-        projectname = project_dirname.split(':')[0]
+        print("Build was added to {0}:".format(builds[0].projectname))
 
-        try:
-            buildopts = buildopts_from_args(args, progress_callback)
-            result = build_function(ownername=username, projectname=projectname,
-                                    project_dirname=project_dirname, buildopts=buildopts, **data)
+        for build in builds:
+            print("  {0}".format(self.build_url(build.id)))
 
-            builds = result if type(result) == list else [result]
-            print("Build was added to {0}:".format(builds[0].projectname))
+        build_ids = [build.id for build in builds]
+        print("Created builds: {0}".format(" ".join(map(str, build_ids))))
 
-            for build in builds:
-                url = urljoin(self.config["copr_url"], "/coprs/build/{0}".format(build.id))
-                print("  {0}".format(url))
+        if not args.nowait:
+            self._watch_builds(build_ids)
 
-            build_ids = [build.id for build in builds]
-            print("Created builds: {0}".format(" ".join(map(str, build_ids))))
 
-            if not args.nowait:
-                self._watch_builds(build_ids)
-
-        except CoprException as ex:
-            sys.stderr.write(str(ex) + "\n")
-            sys.exit(1)
-        finally:
-            if bar:
-                bar.finish()
+    def process_build(self, args, build_function, data):
+        username, projectname, project_dirname = self.parse_dirname(args.copr_repo)
+        buildopts = buildopts_from_args(args)
+        result = build_function(ownername=username,
+                                projectname=projectname,
+                                project_dirname=project_dirname,
+                                buildopts=buildopts, **data)
+        builds = result if type(result) == list else [result]
+        self.print_build_info_and_wait(builds, args)
 
 
     @requires_api_auth
@@ -566,8 +466,14 @@ class Commands(object):
             module_hotfixes=ON_OFF_MAP[args.module_hotfixes],
             fedora_review=args.fedora_review,
             appstream=ON_OFF_MAP[args.appstream],
+            runtime_dependencies=args.runtime_dependencies,
+            packit_forge_projects_allowed=args.packit_forge_projects_allowed,
         )
-        print("New project was successfully created.")
+
+        owner_part = username.replace('@', "g/")
+        project_url = '/'.join([self.client.config["copr_url"].strip("/"),
+                                'coprs', owner_part, copr, ""])
+        print("New project was successfully created: {0}".format(project_url))
 
     @requires_api_auth
     def action_modify_project(self, args):
@@ -592,6 +498,8 @@ class Commands(object):
             module_hotfixes=ON_OFF_MAP[args.module_hotfixes],
             fedora_review=ON_OFF_MAP[args.fedora_review],
             appstream=ON_OFF_MAP[args.appstream],
+            runtime_dependencies=args.runtime_dependencies,
+            packit_forge_projects_allowed=args.packit_forge_projects_allowed,
         )
 
     @requires_api_auth
@@ -681,7 +589,6 @@ class Commands(object):
         print(MockProfile(build_config))
 
 
-    @check_username_presence
     def action_list(self, args):
         """ Method called when the 'list' action has been selected by the
         user.
@@ -689,7 +596,7 @@ class Commands(object):
         :param args: argparse arguments provided by the user
 
         """
-        username = args.username or self.config["username"]
+        username = args.username or self.ownername
         projects = self.client.project_proxy.get_list(username)
         if not projects:
             sys.stderr.write("No copr retrieved for user: '{0}'\n".format(username))
@@ -756,16 +663,27 @@ class Commands(object):
 
         :param args: argparse arguments provided by the user
         """
-
         if args.bootstrap_image:
             args.bootstrap = 'image'
         owner, copr, chroot = self.parse_chroot_path(args.coprchroot)
+
+        with_opts = None
+        if args.rpmbuild_with:
+            with_opts = ' '.join(args.rpmbuild_with)
+
+        without_opts = None
+        if args.rpmbuild_without:
+            without_opts = ' '.join(args.rpmbuild_without)
+
         self.client.project_chroot_proxy.edit(
             ownername=owner, projectname=copr, chrootname=chroot,
             comps=args.upload_comps, delete_comps=args.delete_comps,
             additional_packages=args.packages, additional_repos=args.repos,
-            bootstrap=args.bootstrap, bootstrap_image=args.bootstrap_image,
+            additional_modules=args.modules, with_opts=with_opts,
+            without_opts=without_opts, bootstrap=args.bootstrap,
+            bootstrap_image=args.bootstrap_image,
             isolation=args.isolation,
+            reset_fields=args.reset,
         )
         print("Edit chroot operation was successful.")
 
@@ -779,9 +697,9 @@ class Commands(object):
         project_chroot = self.client.project_chroot_proxy.get(
             ownername=owner, projectname=copr, chrootname=chroot
         )
-        fields = ["additional_packages", "additional_repos", "comps_name", "delete_after_days",
-                  "isolation", "mock_chroot", "ownername", "projectname", "with_opts",
-                  "without_opts"]
+        fields = ["additional_packages", "additional_repos", "additional_modules",
+                  "comps_name", "delete_after_days", "isolation", "mock_chroot",
+                  "ownername", "projectname", "with_opts", "without_opts"]
         printer = get_printer(args.output_format, fields)
         printer.add_data(project_chroot)
         printer.finish()
@@ -810,6 +728,7 @@ class Commands(object):
             "package_name": args.name,
             "pypi_package_name": args.packagename,
             "pypi_package_version": args.packageversion,
+            "spec_generator": args.spec_generator,
             "spec_template": args.spec_template,
             "python_versions": args.pythonversions,
             "max_builds": args.max_builds,
@@ -896,24 +815,46 @@ class Commands(object):
 
     def action_list_packages(self, args):
         ownername, projectname = self.parse_name(args.copr)
-        packages = self.client.package_proxy.get_list(ownername=ownername, projectname=projectname,
-                                                      with_latest_build=args.with_latest_build,
-                                                      with_latest_succeeded_build=args.with_latest_succeeded_build)
-        packages_with_builds = [self._package_with_builds(p, args) for p in packages]
-        fields = ["id", "auto_rebuild", "name", "ownername", "projectname", "source_dict",
-                  "source_type", "latest_succeeded_build", "latest_build"]
+        fields = [
+            "id",
+            "auto_rebuild",
+            "name",
+            "ownername",
+            "projectname",
+            "source_dict",
+            "source_type",
+            "latest_succeeded_build",
+            "latest_build"
+        ]
         if args.with_all_builds:
             fields.insert(1, "builds")
+
+        pagination = {"limit": 1000}
+        packages = self.client.package_proxy.get_list(
+            ownername=ownername, projectname=projectname,
+            with_latest_build=args.with_latest_build,
+            with_latest_succeeded_build=args.with_latest_succeeded_build,
+            pagination=pagination,
+        )
+
         printer = get_printer(args.output_format, fields, True)
-        for data in packages_with_builds:
-            printer.add_data(data)
+        while packages:
+            packages_with_builds = [self._package_with_builds(p, args)
+                                    for p in packages]
+            for data in packages_with_builds:
+                printer.add_data(data)
+            packages = next_page(packages)
         printer.finish()
 
     def action_list_package_names(self, args):
         ownername, projectname = self.parse_name(args.copr)
-        packages = self.client.package_proxy.get_list(ownername=ownername, projectname=projectname)
-        for package in packages:
-            print(package.name)
+        pagination = {"limit": 1000}
+        packages = self.client.package_proxy.get_list(ownername=ownername, projectname=projectname,
+                                                      pagination=pagination)
+        while packages:
+            for package in packages:
+                print(package.name)
+            packages = next_page(packages)
 
     def action_get_package(self, args):
         ownername, projectname = self.parse_name(args.copr)
@@ -965,17 +906,13 @@ class Commands(object):
         print("Package's default source was successfully reseted.")
 
     def action_build_package(self, args):
-        ownername, project_dirname = self.parse_name(args.copr_repo)
-        projectname = project_dirname.split(':')[0]
-        buildopts = buildopts_from_args(args, None)
+        ownername, projectname, project_dirname = self.parse_dirname(args.copr_repo)
+
+        buildopts = buildopts_from_args(args)
         build = self.client.package_proxy.build(ownername=ownername, projectname=projectname,
                                                 packagename=args.name, buildopts=buildopts,
                                                 project_dirname=project_dirname)
-        print("Build was added to {0}.".format(build.projectname))
-        print("Created builds: {0}".format(build.id))
-
-        if not args.nowait:
-            self._watch_builds([build.id])
+        self.print_build_info_and_wait([build], args)
 
     def action_build_module(self, args):
         """
@@ -1030,6 +967,33 @@ class Commands(object):
                                                       request)
         print("success")
 
+
+def create_and_modify_common_opts(parser):
+    """
+    De-duplicate options for 'copr create' and 'copr modify' options.
+    """
+    parser.add_argument(
+        "--repo", dest="repos", action="append", metavar="REPO_BASEURL", help=(
+            "Repository that will be enabled at package build time, can be "
+            "specified multiple times."
+    ))
+    parser.add_argument(
+        "--runtime-repo-dependency", dest="runtime_dependencies",
+        metavar="REPO_BASEURL", action="append", help=(
+            "Repository that will be automatically enabled together "
+            "with repository from this copr, e.g. by `dnf copr enable`. "
+            "This can be specified multiple times."
+    ))
+
+    parser.add_argument(
+        "--packit-forge-project-allowed", dest="packit_forge_projects_allowed",
+        metavar="FORGE_PROJECT", action="append", help=(
+            "Forge project that will be allowed to build in this project "
+            "via Packit in format github.com/packit/ogr. "
+            "Can be specified multiple times."
+    ))
+
+
 def setup_parser():
     """
     Set the main arguments.
@@ -1100,8 +1064,9 @@ def setup_parser():
                                help="The name of the copr to create")
     parser_create.add_argument("--chroot", dest="chroots", action="append",
                                help="Chroot to use for this copr")
-    parser_create.add_argument("--repo", dest="repos", action="append",
-                               help="Repository to add to this copr")
+
+
+
     parser_create.add_argument("--initial-pkgs", dest="initial_pkgs",
                                action="append",
                                help="List of packages URL to build in this "
@@ -1161,6 +1126,8 @@ def setup_parser():
         help=("Generate AppStream metadata for this project. Generating "
               "metadata slows down the builds in large Copr projects."))
 
+    create_and_modify_common_opts(parser_create)
+
     parser_create.set_defaults(func="action_create")
 
     # create the parser for the "modify_project" command
@@ -1173,8 +1140,6 @@ def setup_parser():
                                help="Description of the copr")
     parser_modify.add_argument("--instructions",
                                help="Instructions for the copr")
-    parser_modify.add_argument("--repo", dest="repos", action="append",
-                               help="Repository to add to this copr")
     parser_modify.add_argument("--disable_createrepo", type=str2bool,
                                help="Disable metadata auto generation")
     parser_modify.add_argument("--enable-net", choices=["on", "off"],
@@ -1219,6 +1184,8 @@ def setup_parser():
         help=("Generate AppStream metadata for this project. Generating "
               "metadata slows down the builds in large Copr projects."))
 
+    create_and_modify_common_opts(parser_modify)
+
     parser_modify.set_defaults(func="action_modify_project")
 
     # create the parser for the "delete" command
@@ -1247,8 +1214,7 @@ def setup_parser():
     parser_builds.add_argument("project", help="Which project's builds should be listed.\
                                Can be just a name of the project or even in format\
                                username/project or @groupname/project.")
-    parser_builds.add_argument("--output-format", choices=["text", "json", "text-row"],
-                               help=output_format_help)
+    cli_use_output_format(parser_builds, default=None)
     parser_builds.set_defaults(func="action_list_builds")
 
     #########################################################
@@ -1260,6 +1226,15 @@ def setup_parser():
                                          help="For what Python versions to build (by default: 3 2)")
     parser_pypi_args_optional.add_argument("--packageversion", metavar = "PYPIVERSION",
                                          help="Version of the PyPI package to be built (by default latest)")
+
+    parser_pypi_args_optional.add_argument(
+        "--spec-generator",
+        dest="spec_generator",
+        help="Tool for generating specfile from a PyPI package",
+        choices=["pyp2rpm", "pyp2spec"],
+        default="pyp2rpm",
+    )
+
     parser_pypi_args_optional.add_argument("--template", "-t", dest="spec_template",
                                          help="Spec template to be used to build srpm with pyp2rpm")
 
@@ -1417,11 +1392,6 @@ def setup_parser():
             help="Build packages from SRPM generated by custom script")
     parser_build_custom.set_defaults(func="action_build_custom")
 
-    # create the parser for the "buildfedpkg" command
-    parser_build_distgit = subparsers.add_parser("buildfedpkg", parents=[parser_distgit_args_parent, parser_build_parent],
-                                                  help="DEPRECATED. Use SCM source type instead.")
-    parser_build_distgit.set_defaults(func="action_build_distgit")
-
     # create the parser for the "buildscm" command
     parser_build_scm = subparsers.add_parser("buildscm", parents=[parser_scm_args_parent, parser_build_parent],
                                               help="Builds package from Git/DistGit/SVN repository.")
@@ -1492,7 +1462,13 @@ def setup_parser():
                                       help="space separated string of package names to be added to buildroot")
     parser_edit_chroot.add_argument("--repos",
                                       help="space separated string of additional repo urls for chroot")
-    parser_edit_chroot.add_argument("--isolation", choices=["simple", "nspawn", "default"], default="unchanged",
+    parser_edit_chroot.add_argument("--modules",
+                                      help="comma separated list of modules that will be enabled or disabled in the given chroot, e.g. 'module1:stream,!module2:stream'")
+    parser_edit_chroot.add_argument("--rpmbuild-with", action='append',
+                                      help="rpmbuild --with option, can be set multiple times")
+    parser_edit_chroot.add_argument("--rpmbuild-without", action='append',
+                                      help="rpmbuild --without options, can be set multiple times")
+    parser_edit_chroot.add_argument("--isolation", choices=["simple", "nspawn", "default"],
                                     help="Choose the isolation method for running commands in buildroot.")
 
     parser_edit_chroot.add_argument(
@@ -1510,12 +1486,20 @@ def setup_parser():
         help=("Use a custom container image for initializing Mock's "
               "bootstrap (Implies --bootstrap=image)"))
 
+    parser_edit_chroot.add_argument(
+        "--reset",
+        action="append",
+        help=("Reset this parameters to their respective defaults. "
+              "Possible values are additional_packages, additional_modules, "
+              "isolation, etc. See the output of `copr-cli get-chroot' for all "
+              "the possible field names."),
+    )
+
     parser_edit_chroot.set_defaults(func="action_edit_chroot")
 
     parser_get_chroot = subparsers.add_parser("get-chroot", help="Get chroot of a project")
     parser_get_chroot.add_argument("coprchroot", help="Path to a project chroot as owner/project/chroot or project/chroot")
-    parser_get_chroot.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
-                                   help=output_format_help)
+    cli_use_output_format(parser_get_chroot)
     parser_get_chroot.set_defaults(func="action_get_chroot")
 
     parser_list_chroots = subparsers.add_parser("list-chroots", help="List all currently available chroots.")
@@ -1639,8 +1623,7 @@ def setup_parser():
                                       help="Also display data related to the latest succeeded build for the package.")
     parser_list_packages.add_argument("--with-all-builds", action="store_true",
                                       help="Also display data related to the builds for the package.")
-    parser_list_packages.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
-                                      help=output_format_help)
+    cli_use_output_format(parser_list_packages)
     parser_list_packages.set_defaults(func="action_list_packages")
 
     # package names listing
@@ -1664,8 +1647,7 @@ def setup_parser():
                                     help="Also display data related to the latest succeeded build for each package.")
     parser_get_package.add_argument("--with-all-builds", action="store_true",
                                     help="Also display data related to the builds for each package.")
-    parser_get_package.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
-                                    help=output_format_help)
+    cli_use_output_format(parser_get_package)
     parser_get_package.set_defaults(func="action_get_package")
 
     # package deletion
@@ -1795,18 +1777,34 @@ def setup_parser():
             help=request_help.format('builder'))
     parser_permissions_request.set_defaults(func='action_permissions_request')
 
+    # package monitoring
+    cli_monitor_parser(subparsers)
+
     if argcomplete:
         argcomplete.autocomplete(parser)
     return parser
 
 
-def enable_debug():
-    logging.basicConfig(
-        level=logging.DEBUG,
-        format='[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
-        datefmt='%H:%M:%S'
-    )
-    sys.stderr.write("#  Debug log enabled  #\n")
+def setup_logging(debug):
+    """
+    Configure the global 'log' object so it prints to standard error output
+    (INFO+ messages).  When --debug is used, all logging (even external
+    libraries) is turned on for DEBUG+ level.
+    """
+    if debug:
+        logging.basicConfig(
+            level=logging.DEBUG,
+            format='[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
+            datefmt='%H:%M:%S'
+        )
+        sys.stderr.write("#  Debug log enabled  #\n")
+        return
+
+    stderr = logging.StreamHandler()
+    stderr.setLevel(logging.INFO)
+    formatter = logging.Formatter("%(levelname)s: %(message)s")
+    stderr.setFormatter(formatter)
+    log.addHandler(stderr)
 
 
 def str2bool(v):
@@ -1818,21 +1816,47 @@ def str2bool(v):
         raise argparse.ArgumentTypeError("Boolean value expected.")
 
 
+def _handle_frontend_api_request_error(e, args):
+    if "is not in JSON format" not in str(e):
+        return
+    if not args.debug:
+        sys.stderr.write("Try 'copr-cli --debug' for more info.\n")
+        return
+
+    page_content = e.result.__response__.content
+    if BeautifulSoup:
+        soup = BeautifulSoup(page_content, features="html.parser")
+        page_content = soup.get_text()
+    else:
+        page_content = re.sub(r'<.*?>', '', page_content.decode("utf-8"))
+
+    sys.stderr.write(
+        "\n"
+        "Server response:\n"
+        "----------------\n"
+        "%s\n" % page_content)
+
+
 def main(argv=sys.argv[1:]):
+    # pylint: disable=too-many-branches
     try:
         # Set up parser for global args
         parser = setup_parser()
         # Parse the commandline
         arg = parser.parse_args(argv)
-        if arg.debug:
-            enable_debug()
+        setup_logging(arg.debug)
 
         if not "func" in arg:
             parser.print_help()
             return
 
         commands = Commands(arg.config)
-        getattr(commands, arg.func)(arg)
+        if isinstance(arg.func, str):
+            # Call self method by its name
+            getattr(commands, arg.func)(arg)
+        else:
+            # Call external command method
+            arg.func(commands, arg)
 
     except KeyboardInterrupt:
         sys.stderr.write("\nInterrupted by user.")
@@ -1846,10 +1870,19 @@ def main(argv=sys.argv[1:]):
     except (CoprRequestException, CoprNoResultException) as e:
         sys.stderr.write("\nSomething went wrong:")
         sys.stderr.write("\nError: {0}\n".format(e))
+        _handle_frontend_api_request_error(e, arg)
         sys.exit(1)
     except argparse.ArgumentTypeError as e:
         sys.stderr.write("\nError: {0}\n".format(e))
         sys.exit(2)
+    except CoprConfigException as err:
+        sys.stderr.write("\nError: Copr configuration error. {0}\n".format(err))
+        sys.exit(6)
+    except CoprAuthException as err:
+        sys.stderr.write("\nError: Operation requires API authentication. "
+                         "See the 'AUTHENTICATION' section in man copr-cli(1).\n"
+                         "\nError: {0}\n".format(err))
+        sys.exit(7)
     except CoprException as e:
         sys.stderr.write("\nError: {0}\n".format(e))
         sys.exit(3)
@@ -1857,10 +1890,6 @@ def main(argv=sys.argv[1:]):
         sys.stderr.write("\nError: {0}\n".format(e))
         sys.exit(5)
 
-        # except Exception as e:
-        # print "Error: {0}".format(e)
-        # sys.exit(100)
-
 
 if __name__ == "__main__":
     main()
